
def randomize_array(arr, *, lo=-1.0, hi=1.0, rng=np.random.default_rng()):
    if arr.dtype == float:
        arr[:] = rng.uniform(lo, hi, arr.shape)
    elif arr.dtype == complex:
        arr.real = rng.uniform(lo, hi, arr.shape)
        arr.imag = rng.uniform(lo, hi, arr.shape)
    else:
        raise RuntimeError('dtype must be real or complex')

    
def random_array(dtype, *, shape=None, lo=-1.0, hi=1.0, pad=True, rng=np.random.default_rng()):
    if shape is None:
        shape = random_shape(rng=rng)
            
    if not pad:
        ret = np.empty(shape, dtype=dtype)
    else:
        # FIXME should sometimes make last axis non-contiguous
        shape = np.asarray(shape, dtype=int)
        new_shape = shape + rng.integers(0, 3, size=len(shape))
        ret = np.empty(new_shape, dtype=dtype)

        for d in range(len(shape)):
            if shape[d] < ret.shape[d]:
                ret = np.take(ret, np.arange(shape[d]), axis=d)
                
    assert np.all(np.asarray(shape) == np.asarray(ret.shape))
    randomize_array(ret, lo=lo, hi=hi, rng=rng)
    
    return ret






# FIXME from kszpipe
def estimate_power_spectrum(self, grid, kbin_delim, use_dc=False, use_nyquist=True):
    """
    Estimates auto power spectrum of Fourier-space grid in k-bins.
    Returns (binned_ps, bin_counts), 1-d arrays of length nbins.

    The 'kbin_delim' argument should be an array of length (nbins+1).
    The k-bins for power spectrum estimation are defined by:

           kbin_delim[0] <= k < kbin_delim[1]
              ...
           kbin_delim[nbins-1] <= k < kbin_delim[nbins]
   
    If use_dc=False, then the k=0 mode will not be used.
    If use_nyquist=False, then "Nyquist" modes will not be used.
    """
        
    kbin_delim = np.asarray(kbin_delim, dtype=float)

    assert self.is_fourier_space(grid)
    assert misc_utils.is_sorted(kbin_delim)
    assert kbin_delim[0] >= 0.0
    
    binned_ps, bin_counts = _box_utils.estimate_power_spectrum(grid, self.npix, self.boxsize, kbin_delim, use_dc, use_nyquist)
    return binned_ps, bin_counts


# FIXME from kszpipe
def estimate_power_spectra(self, grids, kbin_delim, faxis, use_dc=False, use_nyquist=True):
    """
    Estimate auto and cross power spectra of multiple Fourier-space grids.

    Returns (binned_ps, bin_counts), where 'binned_ps' has shape (nbins, ngrids, ngrids),
    and 'bin_counts' is a 1-d array of length nbins.

    The 'grids' argument represents a length-ngrids array of Fourier-space grids. The 'faxis'
    argument indicates which axis (usually the first or last) of the 'grids' array is the
    grid index 0 <= i < ngrids.

    The 'kbin_delim' argument should be an array of length (nbins+1).
    The k-bins for power spectrum estimation are defined by:

           kbin_delim[0] <= k < kbin_delim[1]
              ...
           kbin_delim[nbins-1] <= k < kbin_delim[nbins]

    If use_dc=False, then the k=0 mode will not be used.
    If use_nyquist=False, then "Nyquist" modes will not be used.
    """

    # Allows syntax Box.estimate_power_spectra([g1,g2], ..., faxis=0)
    grids = np.asarray(grids)
        
    if (grids.ndim != self.ndim + 1):
        raise RuntimeError("Box.estimate_power_spectra(): expected 'grids' array to have one higher dimension than Box array."
                           + " Maybe you meant to call estimate_power_spectrum() instead of estimate_power_spectra()?")

    assert 0 <= faxis < self.ndim
    grid_shape = grids.shape[:faxis] + grids.shape[(faxis+1):]
    
    if (grid_shape != self.fshape) or (grids.dtype != complex):
        raise RuntimeError(f"Box.estimate_power_spectra(): bad 'grids' shape/dtype (shape={grids.shape}, dtype={grids.dtype}, faxis={faxis})")

    kbin_delim = np.asarray(kbin_delim, dtype=float)
    assert misc_utils.is_sorted(kbin_delim)
    assert kbin_delim[0] >= 0.0
    
    binned_ps, bin_counts = _box_utils.estimate_power_spectra(grids, self.npix, self.boxsize, kbin_delim, faxis, use_dc, use_nyquist)
    return binned_ps, bin_counts


####################################################################################################


def multiply_rfunc(box, arr, f, dest=None):
    pass


def multiply_rhat(box, arr, i, dest=None, eps=1.0e-7):
    pass


def multiply_kfunc(box, arr, f, dest=None):
    """Applies function f(k) to Fourier-space map 'arr', where k = |\vec k| is a scalar wavenumber.

    The 'box' argument is an instance of class kszx.Box.
    The 'arr' argument is a numpy array with shape=box.fourier_space_shape and dtype=complex.
    The 'f' argument can be any callable object k -> f(k), where input/output arrays are multidimensional.

    FIXME arguments for kmin/kmax
    """
    pass


def multiply_khat(box, arr, i, dest=None):
    pass


def apply_partial_derivative(box, arr, d, dest=None):
    """
    Applies partial derivative (\partial / \partial x^d) in place to Fourier-space grid.
    Note that "Nyquist" modes will be zeroed, due to the k/-k ambiguity.
    """

    # FIXME kszpipe code starts here.
    # FIXME here and elsewhere, rename d -> axis.
    
    assert self.is_fourier_space(grid)
    assert 0 <= d < self.ndim
    grid *= (1j * self.get_k_component(d, zero_nyquist=True))


def apply_laplacian(box, arr, dest=None):
    pass


def apply_inverse_laplacian(box, arr, dest=None):
    """
    Applies inverse Laplacian (-1/k^2) in place to Fourier-space grid.
    Note that the k=0 mode will be zeroed, for lack of any better alternative.
    """
    
    # FIXME kszpipe code starts here.
    assert self.is_fourier_space(grid)
    grid /= self.get_k2(multiplier=-1.0, regulate=True)
    grid[(0,)*self.ndim] = 0.0  # Set k=0 to zero


def set_zero_above_kmax(self, grid, kmax):
    """Zeroes modes of Fourier-space grid with (k > kmax)."""
    assert self.is_fourier_space(grid)
    grid *= (self.get_k2() <= kmax**2)
        
    
def project_real(self, grid, *, multiplier=1.0):
    """
    Imposes reality condition f(k) = f(-k)^* on Fourier-space grid.
    The 'multiplier' argument is a hack for convenience in simulate_white_noise().
    """
        
    assert self.is_fourier_space(grid)

    self._project_real(grid[...,0], multiplier)
    if self.npix[-1] % 2 == 0:
        self._project_real(grid[...,-1], multiplier)
        

def _project_real(arr, multiplier):
    """Helper for non-static method project_real()."""
    arr *= (0.5 * multiplier)
    arr += _box_utils.reverse_array(arr).conj()



####################################################################################################


def simulate_gaussian_field(box, pk, kmax=None, zero_dc=False, dest=None):
    """Simulates a Gaussian Fourier-space field with specified power spectrum P(k).

    The 'box' argument is an instance of class kszx.Box.
    The 'pk' argument can be any callable object k -> P(k), where input/output arrays are multidimensional.
    If the 'kmax' argument is specified, then the output array will be zeroed for k > kmax.
    If the 'zero_dc' flag is specified, then 

    Note that the output array is a Fourier-space map (shape box.fourier_space_shape and dtype=complex).
    If a real-space map is preferred, then call fft_c2r() on the output array.

    Reminder: Fourier convention for simulated field is

       <f(k) f(-k')> = (box volume) P(k) delta_{kk'}  [ morally P(k) (2pi)^n delta^n(k-k') ]

    With this convention, calling fft_c2r() on the output gives a real-space map with the correct variance.
    """

    assert isinstance(box, Box)
    assert callable(pk)

    # FIXME kszpipe code starts here (note that pk is a scalar in this code)

    rms = np.sqrt(0.5 * pk * self.box_volume) if normalize else np.sqrt(0.5 * pk)
    ret = np.zeros(self.nk, dtype=complex)        
    ret.real = np.random.normal(size=self.nk, scale=rms)
    ret.imag = np.random.normal(size=self.nk, scale=rms)
    
    if project_real:
        self.project_real(ret, multiplier = np.sqrt(2.))
        
    if not simulate_dc:
        ret[(0,)*self.ndim] = 0.0



import os
import gzip
import shutil
import fitsio

from . import Catalog
from . import io_utils


def read_galaxies(survey, download=False):
    """Returns SDSS galaxy catalog (instance of type kszx.Catalog).

    Example usage:

         gcat = kszx.sdss.read_galaxies('CMASS_North')   # returns kszx.Catalog
         gcat.apply_redshift_cut(0.43, 0.7)              # suggest applying redshift cut
    """
    filename = _galaxy_filename(survey, download)
    return read_fits_catalog(filename, is_randcat=False)


def read_randoms(survey, download=False):
    """Returns SDSS galaxy catalog (instance of type kszx.Catalog).

    Example usage:

         rcat = kszx.sdss.read_randoms('CMASS_North')   # returns kszx.Catalog
         rcat.apply_redshift_cut(0.43, 0.7)             # suggest applying redshift cut
    """
    filenames = _random_filenames(survey, download)
    catalogs = [ read_fits_catalog(f, is_randcat=True) for f in filenames ]
    return Catalog.concatenate(catalogs, name=f'{survey} randoms', destructive=True)


def read_mask(survey, download=False):
    """Not sure what this is used for!

    Returns a pymangle object that can be evaluated with mask.weight(ra_deg, dec_deg).
    The return value is either 0, or close to 1 (e.g. 0.98). For a plot, see
    scripts/sdss_exploratory_plots.ipynb.
    """
    import pymangle
    filename = _mask_filename(survey, download)
    print(f'Reading {filename}')
    return pymangle.Mangle(filename)


def download(survey, mask=False):
    _galaxy_filename(survey, download=True)
    _random_filenames(survey, download=True)
    _mask_filename(survey, download=mask)


####################################################################################################


def read_fits_catalog(filename, is_randcat, name=None, extra_columns=[]):
    """Intended to be called through wrapper such as read_galaxies() or read_randoms()."""

    print(f'Reading {filename}')
    catalog = Catalog(name=name, filename=filename)

    with fitsio.FITS(filename) as f:
        
        # For a list of all fields, see
        #   https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/galaxy_DRX_SAMPLE_NS.html
        #   http://data.sdss3.org/datamodel/files/BOSS_LSS_REDUX/galaxy_DR11v1_SAMPLE_NS.html
        #   http://data.sdss3.org/datamodel/files/BOSS_LSS_REDUX/randomN_DR10v8_SAMPLE_NS.html
        
        catalog.add_column('ra_deg', f[1].read('RA'))
        catalog.add_column('dec_deg', f[1].read('DEC'))
        catalog.add_column('z', f[1].read('Z'))

        # catalog.add_column('ipoly', f[1].read('IPOLY'))
        # catalog.add_column('isect', f[1].read('ISECT'))

        if not is_randcat:
            catalog.add_column('wfkp', f[1].read('WEIGHT_FKP'))
            catalog.add_column('wcp', f[1].read('WEIGHT_CP'))
            catalog.add_column('wzf', f[1].read('WEIGHT_NOZ'))
            catalog.add_column('wsys', f[1].read('WEIGHT_SYSTOT'))
            catalog.add_column('cboss', f[1].read('COMP'))
            catalog.add_column('id', f[1].read('ID'))

        for col_name in extra_columns:
            catalog.add_column(col_name.lower(), f[1].read(col_name.upper()))

    catalog._announce_file_read()
    return catalog


####################################################################################################


def _check_survey(survey):
    """Check that 'survey' is valid, and return its standard capitalization."""
    
    survey_list = [ 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South',
                    'CMASSLOWZTOT_North', 'CMASSLOWZTOT_South', 'CLASSLOWZE2_North',
                    'CMASSLOWZE3_North' ]
                    
    for s in survey_list:
        if survey.upper() == s.upper():
            return s

    raise RuntimeError(f"SDSS survey '{survey}' not recognized (must be one of: {survey_list})")
        
    
def _sdss_path(relpath, download=False, gz=False):
    """Intended to be called through wrapper such as _galaxy_filename(), _random_filenames(), etc.
    If gz=True, then the file is gzipped on the SDSS website (only matters if download=True).
    """
    
    sdss_base_dir = io_utils.get_data_dir('sdss')
    abspath = os.path.join(sdss_base_dir, 'DR12v5', relpath)

    if (not download) or os.path.exists(abspath):
        return abspath

    if not gz:
        url = f'https://data.sdss.org/sas/dr12/boss/lss/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading
        return abspath

    # Case gz=True. Download gzipped file (by calling _sdss_path() recursively) and uncompress.
    
    gzpath = _sdss_path(f'{relpath}.gz', download=True, gz=False)
    print(f'Uncompressing {gzpath}')

    with gzip.open(gzpath, 'rb') as f_in:
        with open(abspath, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)

    return abspath


def _galaxy_filename(survey, download=False):
    s = _check_survey(survey)
    return _sdss_path(f'galaxy_DR12v5_{s}.fits', download, gz=True)


def _random_filenames(survey, download=False):
    s = _check_survey(survey)
    return [ _sdss_path(f'random{n}_DR12v5_{s}.fits', download, gz=True) for n in [0,1] ]


def _mask_filename(survey, download=False):
    s = _check_survey(survey)
    return _sdss_path(f'mask_DR12v5_{s}.ply', download)




import os
import h5py
import fitsio
import pickle   # FIXME cPickle?
import functools
import numpy as np

# Avoid name collision with io_utils.wget() below.
import wget as wget_module


def read_pickle(filename):
    print(f'Reading {filename}')
    with open(filename, 'rb') as f:
        return pickle.load(f)


def write_pickle(filename, x):
    mkdir_containing(filename)
    print(f'Writing {filename}')    
    with open(filename, 'wb') as f:
        pickle.dump(x, f)


def read_npy(filename, verbose=True):
    print(f'Reading {filename}')
    return np.load(filename, allow_pickle=True)


def write_npy(filename, arr, verbose=True):
    mkdir_containing(filename)
    print(f'Writing {filename}')    
    np.save(filename, arr)

    
def read_npz(filename, verbose=True):
    """
    Reminder: returns a dictionary-like object (NpzFile) which maps names to arrays.

    Some arrays may have special names 'arr_0', 'arr_1', ... 
    This happens for arrays which were positional (not keyword) arguments to write_npz().
    """

    # read_npy() and read_npz() can be the same function,
    # since np.load() works for both .npy and .npz files.
    return read_npy(filename, verbose=verbose)


def write_npz(filename, *args, **kwds):
    """
    Example: write_npz('file.npz', a, b, xarr=x, yarr=y), where (a,b,x,y) are numpy arrays.

    In this example, when 'file.npz' is read (with np.load() or kszx.read_npz()), the return 
    value is a dictionary-like object (NpzFile) with keys 'arr_0', 'arr_1', 'xarr', 'yarr'.
    Note that the positional arguments (a,b) get special names 'arr_0', 'arr_1'.
    """

    mkdir_containing(filename)
    print(f'Writing {filename}')    
    np.savez(filename, *args, **kwds)


####################################################################################################


@functools.cache
def get_data_dir(name):
    """
    Usage: get_data_dir('desils') returns

      - the value of the environment variable DESILS_DATA_DIR if defined,

      - otherwise $(root)/deslis, where $(root) is:
          - the value of the environment variable KSZX_ROOT_DATA_DIR if defined,
          - otherwise, /data

    Special case: get_data_dir('kszx_root') returns $(root) as defined above. 
    (Used internally, when calling get_data_dir() recursively.)
    """

    env_varname = f'{name.upper()}_DATA_DIR'
    
    if env_varname in os.environ:
        ret = os.environ[env_varname]
        print(f"Using {name} data directory {ret} from environment variable {env_varname}")
        return ret

    if name.upper() == 'KSZX_ROOT':
        ret = '/data'
    else:
        ret = os.path.join(get_data_dir('kszx_root'), name)
        
    print(f"Using default {name} data directory '{ret}' (set environment variable {env_varname} to override)")
    return ret


def mkdir_containing(filename):
    assert isinstance(filename, str)
    
    d = os.path.dirname(filename)
    if (d == '') or os.path.exists(d):
        return
    
    print(f'Creating directory {d}')
    os.makedirs(d, exist_ok=True)
    

def wget(filename, url):
    if os.path.exists(filename):
        print(f'File {filename} already exists -- skipping download')
        return
    
    mkdir_containing(filename)
    print(f'Downloading {url} -> {filename}')
    
    wget_module.download(url, out=filename)
    print()   # extra newline after progress bar
    assert os.path.exists(filename)


####################################################################################################


def show_file(filename, file_type=None, file=None):
    """Prints (to stdout) a simple human-readable summary of file with given name.

    The file_type can be one of:

      pkl        python pickle
      npy        numpy array
      npz        numpy NpzFile
      h5, hdf5   HDF5 file
      fits       FITS file

    If file_type is None, then the file_type is determined from the filename in 
    a brain-dead way (e.g. if filename='x.pkl', then file_type='pkl').

    Can be invoked from command line with 'python -m kszx show <filename>'.
    """

    assert isinstance(filename, str)
        
    if file_type is None:
        b = os.path.basename(filename)
        i = b.rfind('.')
        if (i < 0) or (i >= len(b)-1):
            raise RuntimeError(f"Couldn't determine file_type from {filename=}")
        file_type = b[(i+1):]

    if file_type == 'pkl':
        x = read_pickle(filename)
        print(x, file=file)

    elif file_type == 'npy':
        x = read_npy(filename)
        print('Shape:', x.shape)
        print(x, file=file)

    elif file_type == 'npz':
        _show_npz(filename, file=file)

    elif file_type in [ 'h5', 'hdf5' ]:
        _show_h5(filename, file=file)

    elif file_type in [ 'fits' ]:
        _show_fits(filename, file=file)
            
    else:
        raise RuntimeError(f"kszx.show_file(): unsupported file_type {file_type}")
        

def _show_npz(x, file=None, indent='  '):
    """Intended to be called through show_file(), or via command line.

    Prints (to stdout) a simple human-readable summary of 'x', which can be
    either a filename or a NpzFile (class numpy.lib.npyio.NpzFile).
    """

    if isinstance(x, str):
        x = read_npz(x)

    if not isinstance(x, np.lib.npyio.NpzFile):
        raise RuntimeError('kszx._show_npz(): argument must be either string or NpzFile')

    print(x, file=file)
    
    for k,v in x.items():
        print(f'{indent}{k}: shape={v.shape}, min={np.min(v)}, max={np.max(v)}', file=file)


def _show_h5_attrs(x, prefix, file):
    for k,v in sorted(x.attrs.items()):
        print(f"{prefix}Attribute '{k}': {v}", file=file)
    

def _show_h5(x, prefix=None, file=None, indent='  '):
    """Intended to be called through show_file(), or via command line.

    Prints (to stdout) a simple human-readable summary of 'x', which can be
    either a filename, an h5py.File, or an h5py.Group.

    The default value of the 'indent' argument indents by two spaces at every level.
    """

    if prefix is None:
        prefix = ''
    if file is None:
        file = sys.stdout
    
    if isinstance(x, str):
        with h5py.File(x) as f:
            print(f"{prefix}File '{x}'", file=file)
            _show_h5(f, prefix + indent, file, indent)
        return

    _show_h5_attrs(x, prefix, file)

    # x.items() = "all groups and datasets in a single dict".
    # First pass: loop over datasets.
    
    for k,v in sorted(x.items()):
        if isinstance(v, h5py.Dataset):
            try:
                s = f', min={np.min(v)}, max={np.max(v)}'
            except:
                s = ''
            print(f"{prefix}Dataset '{k}': shape={v.shape}, dtype={v.dtype}{s}", file=file)
            _show_h5_attrs(v, prefix + indent, file, indent)
        elif not isinstance(v, h5py.Group):
            print(f"{prefix}Unrecognized element '{k}': type={type(v)}", file=file)

    # Second pass: loop over groups

    for k,v in sorted(x.items()):
        if isinstance(v, h5py.Group):
            print(f"{prefix}Group '{k}'", file=file)
            _show_h5(v, prefix + indent, file, indent)


def _show_fits(x, file=None):
    """Intended to be called through show_file(), or via command line.

    Prints (to stdout) a simple human-readable summary of 'x', which can be
    either a filename, or a fitsio.FITS object.

    This quick-and-dirty implementation just calls __str__() functions in the
    fitsio python module (which is pretty good).
    """

    if isinstance(x, str):
        with fitsio.FITS(x) as f:
            _show_fits(f)

    elif isinstance(x, fitsio.FITS):
        for hdu in x:
            print(hdu, file=file)

    else:
        raise RuntimeError('kszx._show_fits(): argument must be a string, or of type fitsio.FITS')



import sys
import argparse


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command')

    p = subparsers.add_parser('download_act')
    p.add_argument('dr', type=int, help='Currently, only dr=5 is allowed')

    p = subparsers.add_parser('download_sdss')
    p.add_argument('survey', help='Survey name such as CMASS_North')

    p = subparsers.add_parser('show')
    p.add_argument('filename')

    p = subparsers.add_parser('test')
    
    args = parser.parse_args()

    if not hasattr(args, 'command'):
        parser.print_help()
        sys.exit(2)
    elif args.command == 'download_act':
        from . import act
        act.download(dr=args.dr)
    elif args.command == 'download_sdss':
        from . import sdss
        sdss.download(args.survey)
    elif args.command == 'show':
        from . import io_utils
        io_utils.show_file(args.filename)
    elif args.command == 'test':
        from . import tests
        tests.run_all_tests()   # defined in kszx/tests/__init__.py
    else:
        parser.print_help()
        sys.exit(2)



import os
import zipfile
import numpy as np
import pixell.enmap

from . import io_utils


def read_cmb(freq, dr, download=False):
    """Returns a pixell map. We currently only support DR5 act_planck_daynight_srcfree maps."""
    
    filename = _cmb_filename(freq, dr, download)
    return _read_map(filename)


def read_ivar(freq, dr, download=False):
    """Returns a pixell map. We currently only support DR5 act_planck_daynight_srcfree maps."""
    
    filename = _ivar_filename(freq, dr, download)
    return _read_map(filename)


def read_beam(freq, dr, lmax=None, download=False):
    """Returns a 1-d numpy array of length (lmax+1). We currently only support DR5 daynight beams."""

    filename = _beam_filename(freq, dr, download)
        
    print(f'Reading {filename}')
            
    a = np.loadtxt(filename)
    assert (a.ndim==2) and (a.shape[1]==2)
    assert np.all(a[:,0] == np.arange(len(a)))

    a = a[:,1]
        
    if lmax is not None:
        assert lmax < len(a)
        a = a[:(lmax+1)]
        
    return a


def read_cluster_mask(download=False):
    filename = _cluster_mask_filename(download)
    return pixell.enmap.read_map(filename)


def read_nilc_wide_mask(download=False):
    filename = _nilc_wide_mask_filename(download)
    return pixell.enmap.read_map(filename)


def download(dr, freq_list=None, cmb=True, ivar=True, beams=True):
    if freq_list is None:
        freq_list = [ 90, 150, 220 ]

    for freq in freq_list:
        _cmb_filename(freq, dr, download=cmb)
        _ivar_filename(freq, dr, download=ivar)
        _beam_filename(freq, dr, download=beams)


####################################################################################################


def _act_path(relpath, dr, download=False, is_aux=False):
    """Intended to be called through wrapper such as _cmb_filename(), _ivar_filename(), etc.

    For files which are downloaded here: https://lambda.gsfc.nasa.gov/data/suborbital/ACT
    If is_aux=True, then the file is contained in 'act_dr5.01_auxilliary.zip' (only matters if download=True)
    """

    assert dr == 5    # currently, only support DR5
    act_base_dir = io_utils.get_data_dir('act')
    abspath = os.path.join(act_base_dir, 'dr5.01', relpath)

    if (not download) or os.path.exists(abspath):
        return abspath

    if not is_aux:
        # Typical case: file is not part of act_dr5.01_auxilliary.zip.
        url = f'https://lambda.gsfc.nasa.gov/data/suborbital/ACT/ACT_dr5/maps/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading
        return abspath

    # Special case: file is part of act_dr5.01_auxilliary.zip.
    # Download the zipfile (by calling _act_path() recursively) and unpack.

    zip_filename = _act_path('act_dr5.01_auxilliary.zip', dr=5, download=True, is_aux=False)
    print(f'Unzipping {zip_filename}')
    
    with zipfile.ZipFile(zip_filename, 'r') as f:
        f.extractall(act_base_dir)

    if not os.path.exists(abspath):
        raise RuntimeError(f"File '{relname}' not found in zip file '{zip_filename}'")
    
    return abspath


def _act_nersc_path(relpath, download=False):
    """Intended to be called through wrapper such as _cluster_mask_filename().
    For files which are downloaded here: https://portal.nersc.gov/project/act
    """

    act_base_dir = io_utils.get_data_dir('act')
    abspath = os.path.join(act_base_dir, relpath)

    if download and not os.path.exists(abspath):
        url = f'https://portal.nersc.gov/project/act/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading

    return abspath
    

def _cmb_filename(freq, dr, download=False):
    assert dr == 5   # currently, only support DR5
    return _act_path(f'act_planck_dr5.01_s08s18_AA_f{freq:03d}_daynight_map_srcfree.fits', dr, download)

def _ivar_filename(freq, dr, download=False):
    assert dr == 5   # currently, only support DR5
    return _act_path(f'act_planck_dr5.01_s08s18_AA_f{freq:03d}_daynight_ivar.fits', dr, download)

def _beam_filename(freq, dr, download=False):
    assert dr == 5   # currently, only support DR5
    return _act_path(f'beams/act_planck_dr5.01_s08s18_f{freq:03d}_daynight_beam.txt', dr, download, is_aux=True)

def _cluster_mask_filename(download=False):
    return _act_nersc_path('dr6_nilc/ymaps_20230220/masks/cluster_mask.fits', download)

def _nilc_wide_mask_filename(download=False):
    return _act_nersc_path('dr6_nilc/ymaps_20230220/masks/wide_mask_GAL070_apod_1.50_deg_wExtended.fits', download)


def _read_map(filename):
    """Called by read_cmb() and read_ivar()."""
    
    print(f'Reading {filename}')

    # FIXME is there a way to avoid reading TQU, if we only want T?
    m = pixell.enmap.read_map(filename)
    assert m.ndim == 3
    
    # Use of .copy() saves memory, by keeping T data while dropping reference to TQU data.
    return m[0].copy()



"""Not much here for now -- just some code to download DR10 randoms. More to come later."""

import os
from . import io_utils


def download_randoms(dr):
    for i in range(20):
        _desils_path(f'randoms/randoms-1-{i}.fits', dr, download=True)


def download_south_randoms(dr):
    for i in range(20):
        _desils_path(f'south/randoms/randoms-south-1-{i}.fits', dr, download=True)

        
def _desils_path(relpath, dr, download=False):
    """Example: _desils_path('randoms/randoms-1-0.fits', dr=10).
    Intended to be called through wrapper, e.g. _random_filename()."""
    
    assert dr == 10    # placeholder for future expansion
    relpath = os.path.join('dr10', relpath)

    desils_base_dir = io_utils.get_data_dir('desils')
    abspath = os.path.join(desils_base_dir, relpath)

    if download and not os.path.exists(abspath):
        url = f'https://portal.nersc.gov/cfs/cosmo/data/legacysurvey/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading
    
    return abspath
    



"""The ``kszx.utils`` module contains miscelleanous utilities, that didn't really fit in elsewhere."""

import numpy as np
import scipy.special
import scipy.integrate
import scipy.interpolate


####################################################################################################


def ra_dec_to_xyz(ra_deg, dec_deg, r=None):
    """Converts spherical polar coordinates (ra_deg, dec_deg) to cartesian coords (x,y,z).
    
    Function arguments:
    
       - ``ra_deg`` (array): RA in degrees (not radians!)
       - ``dec_deg`` (array): DEC in degrees (not radians!)
       - ``r`` (optional array): radial coordinates (units of length)

    Return value:

       - ``xyz`` (array): new array with a length-3 axis appended.
         (E.g. if ra_deg and dec_deg have shape (m,n), then xyz has shape (m,n,3)).

    The coordinate systems are related by::

        x = r * cos(dec) * cos(ra)
        y = r * cos(dec) * sin(ra)
        z = r * sin(dec)
    """
    
    ra_deg = np.asarray(ra_deg)
    dec_deg = np.asarray(dec_deg)
    assert ra_deg.shape == dec_deg.shape

    # FIXME memory-inefficent (uses too many temp arrays)
    
    ra = ra_deg * (np.pi/180.)
    dec = dec_deg * (np.pi/180.)
    cos_dec = np.cos(dec)
    
    ret = np.empty(ra_deg.shape + (3,), dtype=float)
    ret[...,0] = cos_dec * np.cos(ra)
    ret[...,1] = cos_dec * np.sin(ra)
    ret[...,2] = np.sin(dec)

    if r is not None:
        r = np.asarray(r)
        assert r.shape == ra_deg.shape
        ret *= np.reshape(r, r.shape + (1,))
    
    return ret


def xyz_to_ra_dec(xyz, return_r=False):
    """Converts cartesian coords (x,y,z) to spherical polar coordinates (ra_deg, dec_deg).

    Function arguments:
    
       - ``xyz`` (array): array whose last axis has length 3.
       - ``return_r`` (boolean): indicates whether radial coords are returned, see below.

    Return value:

       - ``ra_deg`` (array): RA in degrees (not radians!)
       - ``dec_deg`` (array): DEC in degrees (not radians!)
       - ``r`` (array): radial coordinates (in same units as ``xyz``)
       - If ``return_r`` is True, then the return value is a triple ``(ra_deg, dec_deg, r)``.
         If ``return_r`` is False, then the return value is a pair ``(ra_deg, dec_deg)``.

    The returned arrays have one lower dimension than ``xyz``. (E.g. if ``xyz`` has shape (m,n,3)
    then ``ra_deg``, ``dec_deg``, and ``r`` all have shape (m,n).)

    The coordinate systems are related by::

        x = r * cos(dec) * cos(ra)
        y = r * cos(dec) * sin(ra)
        z = r * sin(dec)
    """

    xyz = np.asarray(xyz)
    assert xyz.shape[-1] == 3

    x = xyz[...,0]
    y = xyz[...,1]
    z = xyz[...,2]

    ra_deg = np.arctan2(y, x)
    ra_deg *= (180./np.pi)
    
    dec_deg = np.arctan2(z, np.sqrt(x*x + y*y))
    dec_deg *= (180./np.pi)

    if return_r:
        r = np.sqrt(x*x + y*y + z*z)
        return ra_deg, dec_deg, r
    else:
        return ra_deg, dec_deg
    

####################################################################################################


def W_tophat(x):
    r"""Returns Fourier transform of a 3-d tophat W(x), where x=kR. Vectorized.
    
    W(x) is given by any of the equivalent forms:

    .. math::

       W(x) &= 3 (\sin(x) - x \cos(x)) / x^3 \\
            &= 3 j_1(x) / x \\
            &= j_0(x) + j_2(x)
    """

    # Timing showed that this implementation was fastest.
    mask = (x != 0)   # to regulate division by zero
    ret = 3 * scipy.special.spherical_jn(1,x) / np.where(mask,x,1.0)
    return np.where(mask, ret, 1.0)


####################################################################################################


def asarray(x, caller, arg, dtype=None, allow_none=False):
    if allow_none and (x is None):
        return x
    
    try:
        assert x is not None
        return np.asarray(x, dtype)
    except:
        pass

    s = f' with dtype {dtype}' if (dtype is not None) else ''
    raise RuntimeError(f"{caller}: couldn't convert {arg} to array{s} (value={x})")
    

def scattered_add(lhs, ix, rhs, normalize_sum = None):
    """Returns sum of RHS values (before applying 'normalize_sum').
    FIXME: Is there a numpy function that does this? If not, write a C++ kernel.
    """
    
    assert lhs.ndim == 1
    assert ix.ndim == 1

    n = len(ix)
    rhs = np.asarray(rhs) if (rhs is not None) else None

    # Split lhs into (scalar, vector) components (the vector can be None).
    if rhs is None:
        rs, rv = 1.0, None
    elif rhs.ndim == 0:
        rs, rv = float(rhs), None
    elif rhs.shape == (n,):
        rs, rv = 1.0, rhs
    else:
        raise RuntimeError("kszx.utils.scattered_add(): 'rhs' and 'ix' have inconsistent shapes")

    rsum = (rs*np.sum(rv)) if (rv is not None) else (rs*n)
    
    if normalize_sum:
        assert rsum > 0.0
        rs *= (normalize_sum / rsum)
        
    # FIXME slow non-vectorized loops, write C++ helper function at some point.
    # (I don't think there is a numpy function that helps.)

    if rv is None:
        for i in range(n):
            lhs[ix[i]] += rs
    elif rs == 1.0:
        for i in range(n):
            lhs[ix[i]] += rv[i]
    else:
        for i in range(n):
            lhs[ix[i]] += rs * rv[i]

    return rsum
            

####################################################################################################


def quad(f, xmin, xmax, epsabs=0.0, epsrel=1.0e-4, points=None):
    """Wrapper for scipy.integrate.quad()."""
    
    assert xmin < xmax

    if points is not None:
        # Remove points outside range (not sure if scipy does this automatically)
        points = [ x for x in points if (xmin < x < xmax) ]                                       
    
    return scipy.integrate.quad(f, xmin, xmax, epsabs=epsabs, epsrel=epsrel)[0]


def dblquad(f, xmin, xmax, ymin, ymax, *, epsabs=0.0, epsrel=1.0e-4):
    """
    Wrapper for scipy.integrate.dblquad().
      - The 'f' argument is f(x,y) [not f(y,x) as in scipy.integrate.dblquad]
      - The 'ymin' and 'ymax' arguments can either be floats or functions of x
    """

    ff = lambda x,y: f(y,x)  # swap
    return scipy.integrate.dblquad(f, xmin, xmax, ymin, ymax, epsabs=epsabs, epsrel=epsrel)[0]


def spline1d(xvec, yvec):
    """Wrapper for scipy.interpolate.InterpolatedUnivariateSpline()."""
    
    assert xvec.shape == yvec.shape
    assert xvec.ndim == yvec.ndim == 1
    return scipy.interpolate.InterpolatedUnivariateSpline(xvec, yvec)


def spline2d(xvec, yvec, zmat):
    """
    Wrapper for scipy.interpolate.RectBivariateSpline().
    
    Note that RectBivariateSpline.__call__(x,y) takes a 'grid' argument:

       grid=False  means "broadcast x,y to get a set of points"
       grid=True   means "take the Cartesian product of x,y to get a grid"

    Warning: grid=True is the default! (I usually want grid=False).
    """

    assert xvec.ndim == yvec.ndim == 1
    assert zmat.shape == (len(xvec), len(yvec))
    return scipy.interpolate.RectBivariateSpline(xvec, yvec, zmat)



def logspace(xmin, xmax, n=None, dlog=None):
    """
    Returns a 1-d array of values, uniformly log-spaced over the range (xmin, xmax).
    
    The spacing can be controlled by setting either the 'n' argument (number of points)
    or 'dlog' (largest allowed spacing in log(x)).
    """

    assert 0 < xmin < xmax
    assert (n is None) or (n >= 2)

    if (n is None) and (dlog is not None):
        n = int((np.log(xmax) - np.log(xmin)) / dlog) + 2
    elif (n is None) or (dlog is not None):
        raise RuntimeError("logspace: exactly one of 'n', 'dlog' should be None")

    ret = np.exp(np.linspace(np.log(xmin), np.log(xmax), n))
    ret[0] = xmin   # get rid of roundoff error
    ret[-1] = xmax  # get rid of roundoff error
    
    return ret


def log_uniform(xmin, xmax, size=None):
    assert 0 < xmin < xmax
    return np.exp(np.random.uniform(np.log(xmin), np.log(xmax), size=size))


def covscale(m):
    m = np.array(m)
    assert m.ndim == 2
    assert m.shape[0] == m.shape[1]
    assert np.all(m.diagonal() > 0.0)

    n = m.shape[0]
    ret = np.zeros((n,n))

    for i in xrange(n):
        for j in xrange(n):
            ret[i,j] = m[i,j] / (m[i,i]*m[j,j])**0.5

    return ret


def is_perfect_square(n):
    if n < 0:
        return False
    m = int(np.round(np.sqrt(n)))
    return n == m**2


def is_sorted(x):
    assert x.ndim == 1
    assert len(x) >= 2
    return np.all(x[:-1] < x[1:])
    

def trapezoid_weights(xmin, xmax, n):
    """Returns xvec, wvec."""

    assert xmin < xmax
    assert n >= 2
    
    xvec = np.linspace(xmin, xmax, n)
    wvec = np.full(n, (xmax-xmin)/(n-1))
    wvec[0] *= 0.5
    wvec[-1] *= 0.5

    return xvec, wvec


def array_slice(arr, axis, *args):
    assert 0 <= axis < arr.ndim
    t1 = (slice(None),) * axis
    t2 = (slice(*args),)
    return a[t1+t2]
        
        
def range_checked_index_operator(source_array, index_array, default_value=None, error_message=None):
    """Returns source_array[index_array], but substitutes 'default_value' if index is out of range."""
    
    assert source_array.ndim == 1
    assert len(source_array) > 0

    n = len(source_array)
    valid = np.logical_and(index_array >= 0, index_array < n)

    if default_value is not None:
        index_array = np.where(valid, index_array, 0)
        return np.where(valid, source_array[index_array], default_value)
    
    if np.all(valid):
        return source_array[index_array]

    if error_message is None:
        error_message = 'index out of range, and no default_value specfied'

    raise RuntimeError(error_message)


def contains_duplicates(l):
    return len(l) != len(set(l))


def boxcar_sum(a, m, normalize=False):
    """
    Given a 1-d array 'a' of length n, return a length-n array containing boxcar-summed values of a.
    Each boxcar consists of (2*m+1) array elements, or fewer near the edges of the array.

    E.g. if m=2, then

      ret[0] = a[0] + a[1] + a[2]
      ret[1] = a[0] + a[1] + a[2] + a[3]
      ret[2] = a[0] + a[1] + a[2] + a[3] + a[4]
      ret[3] = a[1] + a[2] + a[3] + a[4] + a[5]
        ...
      ret[n-4] = a[n-6] + a[n-5] + a[n-4] + a[n-3] + a[n-2]
      ret[n-3] = a[n-5] + a[n-4] + a[n-3] + a[n-2] + a[n-1]
      ret[n-2] = a[n-4] + a[n-3] + a[n-2] + a[n-1]
      ret[n-1] = a[n-3] + a[n-2] + a[n-1]
    
    If 'normalize' is True, then "boxcar means" are computed instead of "boxcar sums".
    """
    
    a = np.asarray(a)
    assert a.ndim == 1
    assert m >= 0

    if m == 0:
        return np.copy(a)
    
    n = len(a)
    assert n > 0
        
    if n <= m+1:
        t = np.mean(a) if normalize else np.sum(a)
        return np.full((n,), t, dtype=a.dtype)

    c = np.cumsum(a)
    ret = np.zeros(n, dtype=a.dtype)
    ret[:-m] = c[m:]
    ret[-m:] = c[-1]
    ret[(m+1):] -= c[:(-m-1)]
    
    if normalize:
        den = np.full((n,), 2*m+1)
        den[:m] -= np.arange(m,0,-1)
        den[-m:] -= np.arange(1,m+1)
        ret = ret / den

    return ret



import kszpipe
import numpy as np


def test_k():
    """Is there any rescuing this disaster??"""
    
    print(f'test_k(): start')
    
    for iouter in range(100):
        box = helpers.random_box()
        exponent = min(np.random.uniform(-2,3), 2)
        regulate = np.random.uniform() < 0.3
        ret = box.get_k(exponent, regulate)

        kmin = min(2*np.pi/box.boxsize) if (regulate or exponent < 0) else 0.
        k2 = [ box.get_k_component(ax, zero_nyquist=False, one_dimensional=True)**2 for ax in range(box.ndim) ]
        epsilon = 0.0
        
        for iinner in range(100):
            ix = [ np.random.randint(0,n) for n in box.nk ]
            if iinner == 0:
                ix = [ 0 for n in box.nk ]

            lhs = ret[np.array(ix)]
            print(f'{ret.shape=} {ix=} {lhs.shape=}')
            rhs = sum(k2[i][j] for i,j in enumerate(ix))
            rhs = max(rhs, kmin)
            rhs = rhs**(exponent/2.)
            print(f'{lhs.shape=} {rhs.shape=}')

            num = np.abs(lhs-rhs) 
            den = (lhs+rhs) if (lhs+rhs > 0) else 1.
            epsilon += (num/den)

        print(f'{epsilon=}')
        
    print(f'test_k(): pass')


####################################################################################################


def test_kmax():
    for iouter in range(100):
        box = kszpipe.Box.make_random()
        if np.all(box.npix == 1):
            continue
        
        kmax = [ np.max(box.get_k_component(d, zero_nyquist=False)) for d in range(box.ndim) ]
        kmax = np.dot(kmax,kmax)**0.5
        eps = np.abs(box.kmax - kmax) / kmax
        assert eps < 1.0e-14
        


####################################################################################################


def make_sinusoid(box, coeff, ivec):
    nd = box.ndim
    assert ivec.shape == (nd,)
    assert ivec.dtype == int

    grid = np.full(box.npix, coeff, dtype=complex)

    for d in range(nd):
        n = box.npix[d]
        u = np.exp((2j*np.pi * ivec[d] / float(n)) * np.arange(n))
        shape = (1,)*d + (n,) + (1,)*(nd-d-1)
        grid *= u.reshape(shape)

    return grid.real


def test_apply_partial_derivative():
    for iouter in range(1000):
        ndim = np.random.randint(1, 4)
        npix = np.random.randint(4, 10, ndim)
        box = kszpipe.Box.make_random(npix)
        coeff = np.random.normal() + np.random.normal()*1j
        ivec = np.random.randint(0, npix)
        d = np.random.randint(0, ndim)
        
        grid = make_sinusoid(box, coeff, ivec)
        grid = box.fft(grid)
        box.apply_partial_derivative(grid, d)
        grid = box.fft(grid)

        n = npix[d]
        i = ivec[d]
        kf = 2*np.pi / box.boxsize[d]
        
        if 2*i < n:
            k = kf * i
        elif 2*i == n:
            k = 0.
        else:
            k = kf * (i-n)

        grid2 = make_sinusoid(box, 1j * k * coeff, ivec)
        eps = np.max(np.abs(grid - grid2)) / kf
        assert eps < 1.0e-13


def test_apply_inverse_laplacian():
    for iouter in range(1000):
        ndim = np.random.randint(1, 4)
        npix = np.random.randint(4, 10, ndim)
        box = kszpipe.Box.make_random(npix)
        coeff = np.random.normal() + np.random.normal()*1j
        ivec = np.random.randint(0, npix)
        d = np.random.randint(0, ndim)

        grid = make_sinusoid(box, coeff, ivec)
        grid = box.fft(grid)
        box.apply_inverse_laplacian(grid)
        grid = box.fft(grid)

        k2 = 0.0
        for d in range(ndim):
            n = npix[d]
            i = ivec[d]
            j = i if (2*i <= n) else (i-n)
            k2 += ((2*np.pi / box.boxsize[d]) * j)**2

        w = (-1.0/k2) if (k2 > 0.0) else 0.0
        grid2 = make_sinusoid(box, w*coeff, ivec)
        
        eps = np.max(np.abs(grid - grid2)) / np.max(box.boxsize)**2
        assert eps < 1.0e-15             
        
        
####################################################################################################


def reverse_array_slow(arr):
    for d in range(arr.ndim):
        n = arr.shape[d]
        ix = [ 0 ] + list(range(n-1,0,-1))
        arr = np.take(arr, ix, axis=d)
    return arr
    

def test_reverse_array():
    for iouter in range(100):
        for dtype in [ float, complex ]:
            arr = kszpipe.misc_utils.random_array(dtype)
            revslow = reverse_array_slow(arr)
            revfast = kszpipe._box_utils.reverse_array(arr)
            assert np.all(revslow == revfast)


####################################################################################################


def rotate_array(arr, roffsets):
    roffsets = np.asarray(roffsets, dtype=int)
    assert np.all(roffsets.shape == (arr.ndim,))

    for d in range(arr.ndim):
        r = roffsets[d]
        n = arr.shape[d]
        assert 0 <= r < n

        s = np.concatenate((np.arange(r,n), np.arange(0,r)))
        arr = np.take(arr, s, axis=d)

    return arr



####################################################################################################
#
# In the following tests:
#   Let P = project_real() operation
#   Let F = fft() in r2c direction
#   Let F^T = fft() in c2r direction
#
# Test 1: F^T F = 1
# Test 2: P^2 = P
# Test 3: PF = F
# Test 4: F F^T P = P
# Test 5: F and F^T are transposes



def test_fft2():
    """Test 2: P^2 = P."""

    for iouter in range(100):
        box = kszpipe.Box.make_random()
        Px = box.simulate_white_noise(fourier=True, normalize=False, project_real=True)
        PPx = np.copy(Px)
        box.project_real(PPx)

        eps = np.max(np.abs(Px - PPx))
        assert eps < 1.0e-13


def test_fft3():
    """Test 3: PF = F."""

    for iouter in range(100):
        box = kszpipe.Box.make_random()
        x = box.simulate_white_noise(fourier=False, normalize=False)
        Fx = box.fft(x)
        
        PFx = np.copy(Fx)
        box.project_real(PFx)

        eps = np.max(np.abs(Fx-PFx)) / np.max(np.abs(Fx))
        assert eps < 1.0e-13


def test_fft4():
    """Test 4: F F^T P = P."""

    for iouter in range(100):
        box = kszpipe.Box.make_random()
        Px = box.simulate_white_noise(fourier=True, normalize=False, project_real=True)
        FPx = box.fft(Px)
        FFPx = box.fft(FPx)

        eps = np.max(np.abs(Px - FFPx))
        assert eps < 1.0e-13


####################################################################################################


def test_white_noise():
    if False:
        print(f'Note: test_white_noise() is a little slow (~10 seconds) and currently disabled')
        return
        
    for npix in [ [4,4], [4,5], [5,4], [5,5] ]:
        box = kszpipe.Box.make_random(npix=npix)
        pk = np.random.uniform(10.0, 100.0)

        acc_shape = box.fshape + (2,)
        acc = np.zeros(acc_shape)
        acc2 = np.zeros(acc_shape)
    
        for nmc in range(1,20001):
            grid_f = box.simulate_white_noise(fourier=True, pk=pk)
            grid_r = box.simulate_white_noise(fourier=False, pk=pk)
            grid_f2 = box.fft(grid_r)

            acc[...,0] += grid_f.real**2
            acc[...,1] += grid_f.imag**2
            acc2[...,0] += grid_f2.real**2
            acc2[...,1] += grid_f2.imag**2

        var = 0.5 * pk * box.box_volume
        eps = np.max(np.abs(acc-acc2)) / nmc / var
        print(f'test_white_noise: {box}, eps={eps}')
        assert eps < 0.2

        
def test_white_noise2():
    if False:
        print(f'Note: test_white_noise2() is a little slow (~10 seconds) and currently disabled')
        return

    for npix in [ [16,16], [16,15], [15,16], [15,15] ]:
        box = kszpipe.Box.make_random(npix=npix)
        pk = np.random.uniform(10.0, 100.0)
        nbins = 5
    
        kbin_delim = np.linspace(0.0, 1.01 * box.kmax, nbins+1)
        accum_ps = np.zeros(nbins)
    
        for nmc in range(1,10001):
            grid = box.simulate_white_noise(fourier=True, pk=pk)
            ps, counts = box.estimate_power_spectrum(grid, kbin_delim, use_dc=True, use_nyquist=True)
            assert np.sum(counts) == np.prod(box.npix)
            accum_ps += ps

        eps = np.max(np.abs(accum_ps/pk/nmc) - 1.0)
        print(f'test_white_noise2: {box}, eps={eps}')
        assert eps < 0.01


####################################################################################################

    
def test_box_filter():
    rng = np.random.default_rng()
    
    for iouter in range(100):
        box = kszpipe.Box.make_random()
        a = np.min(box.boxsize) * rng.uniform(-1.0, 1.0)
        b = rng.uniform(0, 2*np.pi)
        f = lambda k: np.sin(a*k + b)
        dc = rng.uniform() if (rng.uniform() > 0.5) else None
        bf = kszpipe.BoxFilter(box, f, dc=dc)

        grid = box.simulate_white_noise(fourier=True, normalize=False)
        
        fgrid = np.copy(grid)
        bf.apply(fgrid)

        eps = 0.0
        
        for i in generate_indices(box.fshape):
            k = np.array(i, dtype=float)
            k = np.minimum(k, box.npix - k)
            k *= (2*np.pi) / box.boxsize
            k = np.dot(k,k)**0.5
            fk = f(k) if ((k > 0) or (dc is None)) else dc
            fg2 = fk * grid[i]
            eps = max(eps, np.abs(fg2 - fgrid[i]))

        assert eps < 1.0e-13
